package com.novell.oauth.samples;

import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.nio.file.FileSystems;
import java.security.spec.RSAPrivateCrtKeySpec;
import java.util.List;
import java.util.Optional;
import java.util.Scanner;

import org.apache.commons.io.IOUtils;
import org.jose4j.jwa.AlgorithmConstraints;
import org.jose4j.jwe.JsonWebEncryption;
import org.jose4j.jwk.JsonWebKey;
import org.jose4j.jwk.JsonWebKeySet;
import org.jose4j.jwk.RsaJsonWebKey;
import org.jose4j.jwk.VerificationJwkSelector;
import org.jose4j.jws.JsonWebSignature;
import org.jose4j.lang.JoseException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class TokenDecoder
{

	private static final Logger LOGGER = LoggerFactory.getLogger(TokenDecoder.class);
	public static List<JsonWebKey> readKeySet(String jwksFileName) throws IOException, JoseException{
		String jwks;
		try(ByteArrayOutputStream baos = new ByteArrayOutputStream();
				FileInputStream fin = new FileInputStream(jwksFileName)){
				IOUtils.copy(fin, baos);
				jwks = baos.toString("UTF-8");
			}
			JsonWebKeySet jsonWebKeySet = new JsonWebKeySet(jwks);
			return jsonWebKeySet.getJsonWebKeys();
	}
	
	public static void displayIDTokenWhenSignAlgNONE(String jwt) throws JoseException{
		
		JsonWebSignature jws = new JsonWebSignature();
	    jws.setCompactSerialization(jwt);
	    jws.setAlgorithmConstraints(AlgorithmConstraints.ALLOW_ONLY_NONE);
	    System.out.println(String.format("%s: %s","Header ",jws.getHeaders().getFullHeaderAsJsonString()));
	    String payload = jws.getPayload();
	    //String payload = jws.getUnverifiedPayload(); //skips singature verification
	    System.out.println(String.format("%s:%s", "Token Payload" , payload));
	}
	
	public static void displayIdTokenIfSignatureMatches(String jwt, String jwksFileName) throws JoseException, MalformedURLException, IOException{
		
		List<JsonWebKey> jsonWebKeys = readKeySet(jwksFileName);
	
		
		JsonWebSignature jws = new JsonWebSignature();
	    
	    jws.setCompactSerialization(jwt);
	    
	    System.out.println(String.format("%s: %s","Header ",jws.getHeaders().getFullHeaderAsJsonString()));
	    LOGGER.info(String.format("%s:%s","Kid in JWT Token", jws.getKeyIdHeaderValue()));
	    if ( null == jsonWebKeys) {
	    	 LOGGER.info(String.format("%s : %s", "Invalid JWKS or unable "
	    			+ "to read JWKS from file", jwksFileName ));
			return ;
	    }
		JsonWebKey jsonWebkey = jsonWebKeys.stream().filter( p -> 
		p.getKeyId() != null && p.getKeyId().equalsIgnoreCase(jws.getKeyIdHeaderValue())).findFirst().orElse(null);
		if(jsonWebkey == null){
			 LOGGER.info("Key ID does not match or not exists, use the first key from JWKS file");
			jsonWebkey= jsonWebKeys.get(0);
		} 
		if(jsonWebkey == null){
			 LOGGER.info("Could not find json web key in JWKS file." );
			return;
		}
		
		jws.setKey(jsonWebkey.getKey());
	    if(false == jws.verifySignature()){
	    	LOGGER.info("signature mismatch");
	    	return;
	    }

	    System.out.println("Signature verified successfully");
	    String payload = jws.getPayload();
	    System.out.println(String.format("%s:%s", "Token Payload" , payload));
	    
	}
	public static String getIDTokenIfDecryptionSuccessful(String jwt,String jwksFileName) throws IOException, JoseException{
		
		JsonWebEncryption jwe = new JsonWebEncryption();
        jwe.setCompactSerialization(jwt);
        System.out.println(String.format("%s: %s","Header ",jwe.getHeaders().getFullHeaderAsJsonString()));
        LOGGER.info("Kid in JWT : "+ jwe.getKeyIdHeaderValue());
		List<JsonWebKey> jsonWebKeys = readKeySet(jwksFileName);
		
		if ( null == jsonWebKeys) {
			LOGGER.info(String.format("%s : %s", "Invalid JWKS or unable "
	    			+ "to read JWKS from file", jwksFileName ));
			return null;
	    }
		
		JsonWebKey jsonWebkey = (JsonWebKey) jsonWebKeys.stream().filter( p -> 
		p.getKeyId() != null && p.getKeyId().equalsIgnoreCase(jwe.getKeyIdHeaderValue())).findFirst().orElse(null);
		if(jsonWebkey == null){
			LOGGER.info("Encryption Key ID does not match or not exists, use the first key from JWKS file");
			jsonWebkey= jsonWebKeys.get(0);
		} 
		if(jsonWebkey == null){
			 LOGGER.info("Could not find json web key in JWKS file." );
			return null;
		}
		
		LOGGER.info("Using Kid to decrypt:  "+ jsonWebkey.getKeyId());
        if(jsonWebkey instanceof RsaJsonWebKey){
        	RsaJsonWebKey rsa = (RsaJsonWebKey) jsonWebkey;
        	jwe.setKey(rsa.getPrivateKey());
        }else
        {
        	jwe.setKey(jsonWebkey.getKey());
        }

       
        String decrString = jwe.getPlaintextString();
        
        return decrString;
 	}
	public static void displayMainOptions()
	{
		System.out.println("option a: plain");
		System.out.println("option b:verify signature");
		System.out.println("option c:decryptjwt");
		System.out.println("option d:DecryptAndVerifySignature");
		System.out.println("option e:Exit");
	}
	public static void main (String [] args){
		Scanner scanner = null;
		try
		{
			scanner = new Scanner(System.in);
			
			System.out.print("Enter jwt token:");
			String jwt = scanner.next().trim();
			displayMainOptions();
			System.out.println("Choose option:");
			String input = scanner.next().trim();
			switch(input.charAt(0))
			{
				case 'a':
							{
								displayIDTokenWhenSignAlgNONE(jwt);
								break;
							}
				case 'b' :
							{
									System.out.print("Enter IDP's JWKS file path:");
									String jwksFilePath = scanner.next();
									displayIdTokenIfSignatureMatches(jwt,jwksFilePath);
									break;
							}
				case 'c' :
							{
								System.out.print("Enter the client's JWKS file path:");
								String jwksFilePath = scanner.next();
								String idToken = getIDTokenIfDecryptionSuccessful(jwt,jwksFilePath);
								System.out.println(idToken);
								break;
								
							}
				case 'd' :
							{
								
								System.out.print("Enter the encryption key's JWKS file path:");
								String clientFilePath =scanner.next();//E:\\Sangeetha\\workspace\\NAM\\JWTUtilityTool\\JWTUtilityTool\\jwksEncryptionKeys.txt 
																		//E:\\Sangeetha\\workspace\\NAM\\JWTUtilityTool\\JWTUtilityTool\\jwksSymmetricKey.txt
								System.out.print("Enter OAuth Provider(IDP)'s Signing Key JWKS file path:");
								String idpFilePath =scanner.next(); //E:\\Sangeetha\\workspace\\NAM\\JWTUtilityTool\\JWTUtilityTool\\jwksSigningKey.txt 
								String idToken = getIDTokenIfDecryptionSuccessful(jwt,clientFilePath);
								System.out.println("Decrypted jwt is :" + idToken);
								LOGGER.info("Verifying the JWT signature...");
								displayIdTokenIfSignatureMatches(idToken,idpFilePath);	
								break;
							}
				case 'e':
							{
								return;
							}
				default:
							{
								System.out.println("invalid option");
							}
							
			}
		}
		catch(Exception e){
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
	}
}
